"""
Base update command and utils
"""
from typing import Any

from pynamodb.models import Model as PynamoModel

from wavelength.dal.dal_config import DalConfig
from wavelength.dal.dao.commands.post_filter_pynamo_command import PostFilteredPynamoCommand
from wavelength.dal.pynamo_models.util import pynamo_write_handler
from wavelength.errors.exceptions import Base404Exception
from wavelength.model.util import get_posix_timestamp


class BaseUpdatePynamoCommand(PostFilteredPynamoCommand):
    """
    Base functionality for updating a pynamo record
    """

    def _execute(self) -> Any:
        """
        Run the command
        :return:Union[List[PynamoPersistenceModel],PynamoPersistenceModel]
        """
        result = super()._execute()
        if result is None:
            raise Base404Exception(
                'Model does not exist', 'Unable to perform update on None')
        return self._update_record(result)

    @pynamo_write_handler
    def _update_record(self, record: PynamoModel):
        """
        Updates the validated record based on the parent model's change set
        :param record: PynamoModel
        :return: None
        """
        next_version = record.version + 1
        model_actions = self._build_model_update_actions()
        default_actions = {
            'table_state': self._parent.db_model.table_state.set(DalConfig.table_state_modified),
        }
        system_actions = {
            'version': self._parent.db_model.version.set(next_version),
            'modified_at': self._parent.db_model.modified_at.set(get_posix_timestamp())
        }
        actions = list({**default_actions, **model_actions, **system_actions}.values())  # allow overrides

        record.update(
            actions=actions,
            condition=(self._parent.db_model.version == record.version)
        )
        self._parent.override_protected_field('version', next_version)
        self._parent.flush_changes()



"""
Lambda log wrapper
"""
import functools

from wavelength.aws.aws_args import convert_response, convert_body
from wavelength.errors.exceptions import (Base5xxException,
                                                   WAVELENGTHBaseException)
from wavelength.logging.log_lambda_base import LogLambdaBase
from wavelength.logging.log_lambda_metrics import LogLambdaMetrics
from wavelength.logging.logging_util import FluentLogger
from wavelength.logging.util.cast_checks import is_json_serializable


class LogLambda(LogLambdaBase):
    """
    This is a decorator that will record the invoking of a lambda handler
    and its input event.

    Also, it will wrap the handler to record the exceptions in a structured
    format for ingestion into the log monitor es cluster.

    the decorator should be initialized with the logger returned
    from _get_structured_logger()
    """

    @staticmethod
    def _flush_buffer(**kwargs):
        return_status = kwargs.get('return_status')

        if return_status:
            if is_json_serializable(return_status):
                LogLambda.info('LogLambda#return_status', 'return_status', data=return_status)
            else:
                LogLambda.warn('LogLambda#_flush_buffer', 'provided return state is not json serializable')

        LogLambda.info('LogLambda#metrics', 'metrics', data=LogLambdaMetrics.sanitized())
        LogLambdaMetrics.reset_metrics()

    def __call__(self, lambda_func):
        """
        If there are decorator arguments, __call__() is only called
        once, as part of the decoration process! You can only give
        it a single argument, which is the function object.
        """
        LogLambdaMetrics.reset_metrics()

        @functools.wraps(lambda_func)
        def log_decorator(*args, **kwargs):
            """
            Decorator function call to setup structured logger
            """
            event = kwargs.get('event', args[0])
            if event:
                event['body'] = convert_body(event.get('body'))

            # Setting the context here, it is safe to assume a 'global'
            # context setting will have no conflict since this is lambda
            # runtime specific
            context = kwargs.get('context', args[1])
            # context.structured_logger = self._set_structured_logger(context, lambda_func.__name__)
            context.structured_logger = self._set_structured_logger(
                context, logger_name=lambda_func.__name__, service_tag=self.service_tag)

            # TODO we can evaluate the entry point for an api gateway context and limit length based on its presence.
            # Currently using simplest path of removing log limit on lambda entry point to reduce overhead of
            # processing input event
            if event and event.get('source') == 'serverless-plugin-warmup':
                # Skipping serverless plugin logging
                self.debug('TRACE',
                           'serverless plugin warmup invocation, skipping processing',
                           state='Skip', **kwargs)
                self._clear_structured_logger()
                return None

            self.info('TRACE', event, state='Invoked', **kwargs)
            try:
                # Capture code for serverless warmup plugin
                ret = lambda_func(*args, **kwargs)
            except Base5xxException as error:
                self.exception(error)
                ret = error.get_response(context)
                raise error
            except WAVELENGTHBaseException as error:
                self.log_wavelength_exception(error)
                ret = error.get_response(context)
            except Exception as error:
                self.exception(error)
                ret = str(error)
                raise Base5xxException(str(error)) from error
            finally:
                self.info('TRACE', ret, state='Completed', **kwargs)
                self._flush_buffer(return_status=ret)
                self._clear_structured_logger()
            return convert_response(ret)

        return log_decorator


class StructLog:
    """
    Structured logger wrapper, emulates log type functionality
    """

    def __init__(self, **value):
        """
        Initializer allows for extra binds
        """
        self.bind(**value)

    @classmethod
    def bind(cls, **value):
        """
        add binding arguments to structured logger
        """
        LogLambda.bind(**value)

    @staticmethod
    def start_timer(name):
        """
        Wrapper to LogLambdaMetrics.start_timer
        """
        LogLambdaMetrics.start_timer(name)

    @staticmethod
    def stop_timer(name):
        """
        Wrapper to LogLambdaMetrics.stop_timer
        """
        LogLambdaMetrics.stop_timer(name)

    @staticmethod
    def counter(name: str, increment: int) -> None:
        """
        Wrapper to LogLambdaMetrics.counter
        """
        LogLambdaMetrics.counter(name, increment)

    @staticmethod
    def gauge(name: str, item) -> None:
        """
        Wrapper to LogLambdaMetrics.gauge
        """
        LogLambdaMetrics.gauge(name, item)

    @staticmethod
    def sets(name: str, item) -> None:
        """
        Wrapper to LogLambdaMetrics.sets
        """
        LogLambdaMetrics.sets(name, item)

    @classmethod
    def exception(cls, error, limit_output=False, **kwargs):
        """
        log exception to structure logger
        """
        if 'reason' not in kwargs:
            if isinstance(error, WAVELENGTHBaseException):
                reason = error.reason
            else:
                reason = 'unknown'
        else:
            reason = kwargs.pop('reason')

        return cls.error('EXCEPTION',
                         str(error),
                         error_type=type(error).__name__,
                         exc_info=True,
                         limit_output=limit_output,
                         reason=reason,
                         **kwargs)

    @classmethod
    def critical(cls, event, msg, limit_output=True, **kwargs):
        """log critical to structure logger - wrapper to _log classmethod"""
        return cls._process_log("critical", event, msg,
                                limit_output=limit_output,
                                **kwargs)

    @classmethod
    def error(cls, event, msg, limit_output=False, **kwargs):
        """log error to structure logger - wrapper to _log classmethod"""
        return cls._process_log("error", event, msg,
                                limit_output=limit_output,
                                **kwargs)

    @classmethod
    def info(cls, event, msg, limit_output=True, **kwargs):
        """log info to structure logger - wrapper to _log classmethod"""
        return cls._process_log("info", event, msg,
                                limit_output=limit_output,
                                **kwargs)

    @classmethod
    def debug(cls, event, msg, limit_output=True, **kwargs):
        """log debug to structure logger - wrapper to _log classmethod"""
        return cls._process_log("debug", event, msg,
                                limit_output=limit_output,
                                **kwargs)

    @classmethod
    def warn(cls, event, msg, limit_output=True, **kwargs):
        """log warn to structure logger - wrapper to _log classmethod"""
        return cls._process_log("warning", event, msg,
                                limit_output=limit_output,
                                **kwargs)

    @classmethod
    def _process_log(cls, log_type, event, msg, **kwargs):
        """
        General log processor, calls the structured log process
        """
        fluent_logger = kwargs.get('fluent_logger')
        LogLambda.process_log(log_type, event, msg, **kwargs)
        if not fluent_logger:
            return FluentLogger(cls, event, msg, **kwargs)
        return None



"""
Pynamo Attribute for supporting compressed JSON
"""

# TODO move to lib

import base64
import zlib
from typing import Optional

import simplejson as json
from pynamodb.attributes import JSONAttribute

from wavelength.logging.slog import StructLog as log

COMPRESSED_KEY = '__COMP__'


class CompressedJSONAttribute(JSONAttribute):
    """
    Compressed nested document that is serialized/deserialized from Dynamo
    """

    def serialize(self, value: dict) -> str:
        """
        Takes a json-serializable object and compresses it
        :param value: dict to serialize
        :return:str
        """

        return super().serialize(self._zip(value))

    def deserialize(self, value: str) -> Optional[dict]:
        """
        Deserializes a JSON string and unpacks it into normal dict if compressed, otherwise returns raw
        :param value: str
        :return: dict
        """
        if not value:
            return None

        return self._unzip(super().deserialize(value))

    @classmethod
    def _zip(cls, source: dict) -> dict:
        """
        serialize and encode a dict,  then decode as ascii to reduce char set size
        :param source:
        :return:
        """

        if not source or not isinstance(source, dict):
            return source

        result = base64.b64encode(
            zlib.compress(
                json.dumps(source).encode()
            )
        ).decode('ascii')  # reduce charset

        return {
            COMPRESSED_KEY: result
        }

    @classmethod
    def _unzip(cls, source: dict) -> dict:
        """
        check source dict for compression key and decompress if present, otherwise pass through
        :param source: dict
        :return: dict
        """

        if not source or not isinstance(source, dict):
            return source

        if COMPRESSED_KEY not in source:
            return source

        try:
            return json.loads(zlib.decompress(base64.b64decode(source[COMPRESSED_KEY])).decode())
        except ValueError as verr:
            log.warn('Pynamo Nested Decompression Failure', source, error=verr)
            return source



"""
Standard response formatter for Lambda
"""


def get_standard_response(body, status_code=200, headers=None, add_cors=True):
    """
    Creates the standard response for a lambda return.
    params
        body: a string or object to return, do not covert using json dumps before,
               this will be handled in the log lambda wrapper to enahnce return
               logging by keeping the response as an object until it is returned.
        status_code: standard HTTP status code result, use the standard exception to handle non 200 and 300 results.
                     Do not handle 400 or 500 errors here. Raise and standard 400 or 500 exception will be handled by
                     the log lambda decorator
        headers: Any additional non CORS header, for CORS header use the add_cors parameter (true by default_
        add_cors: Adds cors headers, default is true (always)
    """
    output, out_headers = {'statusCode': status_code}, {}
    if headers:
        out_headers.update(headers)
    if add_cors:
        out_headers.update({
            "Access-Control-Allow-Origin": "*",
            "Access-Control-Allow-Credentials": True
        })
    output['headers'] = out_headers
    output['body'] = body
    return output



"""

This class much like the LogLambdaBase class, is intended to act as a singleton. Whats more is that it is intended
to work in tandem with the existing logger codebase. Although a completely separate entity, this class and the metrics
it generates is to be attached to the log buffer before it is json seralized and flushed. As a result the following code
protects against adding objects to the metrics dictionary that are not json serializable.
"""

import datetime
from wavelength.logging.log_lambda_base import LogLambdaBase
from wavelength.logging.util.cast_checks import is_int_like, is_json_serializable


class LogLambdaMetrics:
    """
    This class is a singleton that has functionality that writes metrics gathering information into a dictionary.
    """
    KEY_COUNTERS = 'counters'
    KEY_TIMERS = 'timers'
    KEY_GAUGES = 'gauges'
    KEY_SETS = 'sets'

    metrics = {}

    @staticmethod
    def reset_metrics():
        """
        reset_metrics will as the name suggests re-set the metrics attribute to the default value.
        """
        LogLambdaMetrics.metrics = {
            LogLambdaMetrics.KEY_COUNTERS: {},
            LogLambdaMetrics.KEY_TIMERS: {},
            LogLambdaMetrics.KEY_GAUGES: {},
            LogLambdaMetrics.KEY_SETS: {},
        }

    @staticmethod
    def start_timer(name):
        """
        start_timer provided a specific name will start the process for a measuring time.
        The caller must also be responsible for calling stop_timer.
        """
        if not LogLambdaMetrics.metrics:
            return

        LogLambdaMetrics.metrics[LogLambdaMetrics.KEY_TIMERS][str(name)] = (datetime.datetime.now())

    @staticmethod
    def stop_timer(name):
        """
        stop_timer when provided a name will attempt to stop the timer namespaced to that name. It assumes that
        start_timer has been called with that same name param. And therefore it expects an instance of
        datetime.datetime to be at that location. This is replaced with an float of the time elapsed since start.
        returns: None
        """
        if not LogLambdaMetrics.metrics:
            return

        attr = str(name)
        if attr not in LogLambdaMetrics.metrics[LogLambdaMetrics.KEY_TIMERS]:
            LogLambdaBase.warn('LogLambdaMetrics#stop_timer',
                               ("stop_timer called on non-existing timer - %s" % name))
            return

        start_time = LogLambdaMetrics.metrics[LogLambdaMetrics.KEY_TIMERS][attr]
        if isinstance(start_time, datetime.datetime):  # meaning it has not been set
            LogLambdaMetrics.metrics[LogLambdaMetrics.KEY_TIMERS][attr] = (
                datetime.datetime.now() - start_time).total_seconds()
        else:
            LogLambdaMetrics.metrics[LogLambdaMetrics.KEY_TIMERS].pop(attr)
            LogLambdaBase.warn(
                'LogLambdaMetrics#stop_timer',
                ("stop_timer encountered something other than a datetime.datetime instance - %s" % name))

    @staticmethod
    def counter(name: str, increment: int) -> None:
        """
        counter when procided a name and an increment will do one of 2 thigns. It will either create a counter object
        with the inputted value, or increment the pre-exisitng value by the increment param.
        returns: None
        """
        if not LogLambdaMetrics.metrics:
            return

        if not is_int_like(increment):
            LogLambdaBase.warn('LogLambdaMetrics#counter', 'called with object that is not int castable')
            return

        name, increment = str(name), int(increment)
        if name not in LogLambdaMetrics.metrics[LogLambdaMetrics.KEY_COUNTERS]:
            LogLambdaMetrics.metrics[LogLambdaMetrics.KEY_COUNTERS][name] = increment
            return

        LogLambdaMetrics.metrics[LogLambdaMetrics.KEY_COUNTERS][name] += increment

    @staticmethod
    def gauge(name: str, item) -> None:
        """
        gauge simply sets the item to the provided namespace. Caller must ensure
        that the items placed into sets are json serializable. This is because
        we convert logs to json later down the line.
        """
        if not LogLambdaMetrics.metrics:
            return

        if not is_json_serializable(item):
            LogLambdaBase.warn('LogLambdaMetrics#gauge', 'called with object that is not json serializable')
            return

        LogLambdaMetrics.metrics[LogLambdaMetrics.KEY_GAUGES][str(name)] = item

    @staticmethod
    def sets(name: str, item) -> None:
        """
        set etiher creates a set with the inputted item or appends the new item if a set already exists.
        """
        if not is_json_serializable(item):
            LogLambdaBase.warn('LogLambdaMetrics#sets', 'called with object that is not json serializable')
            return

        if LogLambdaMetrics.metrics:
            if name in LogLambdaMetrics.metrics[LogLambdaMetrics.KEY_SETS]:
                LogLambdaMetrics.metrics[LogLambdaMetrics.KEY_SETS][name].add(item)
            else:
                LogLambdaMetrics.metrics[LogLambdaMetrics.KEY_SETS][name] = {item}

    @staticmethod
    def sanitize_timestamps(timestamps: {}) -> dict:
        """
        sanitize_timestamps ensures that there are no lingering instances of the datetime class. There helps protect
        against the case where a user has started a timestamp but has failed to close it.
        """
        new_dictionary = {}
        for k, value in timestamps.items():
            if not isinstance(value, datetime.datetime):
                new_dictionary[k] = value
        return new_dictionary

    @staticmethod
    def sanitize_sets(sets: {}) -> dict:
        """
        sanitize_sets converts the sets attribute into lists for the purposes of json seralization.
        """
        for k, value in sets.items():
            sets[k] = list(value)

        return sets

    @staticmethod
    def sanitized():
        """
        sanitize calls the other sanitize methods.
        """
        if LogLambdaMetrics.metrics:
            LogLambdaMetrics.metrics[LogLambdaMetrics.KEY_TIMERS] = LogLambdaMetrics.sanitize_timestamps(
                LogLambdaMetrics.metrics.get(LogLambdaMetrics.KEY_TIMERS))
            LogLambdaMetrics.metrics[LogLambdaMetrics.KEY_SETS] = LogLambdaMetrics.sanitize_sets(
                LogLambdaMetrics.metrics.get(LogLambdaMetrics.KEY_SETS))
        return LogLambdaMetrics.metrics



"""
Validation utils for Bravado
"""
from bravado_core.request import unmarshal_request
from bravado_core.spec import Spec
from bravado_core.validate import validate_object
from jsonschema import ValidationError
from schleppy import reach
from yaml import load, Loader

from wavelength.errors.exceptions import Base422Exception
from wavelength.validation.util import RequestValidationModel


class Validator:
    """
    Validation helper class for validating against OpenAPI schemas
    """

    def __init__(self, spec_file_path, options: dict = None):
        if not options:
            options = {
                'validate_swagger_spec': False,
                'validate_requests': True,
                'validate_responses': False,
                'use_models': True,
            }
        with open(spec_file_path, 'r') as spec:
            self.spec_source = load(spec.read(), Loader)
            self.spec = Spec.from_dict(self.spec_source, config=options)

    def validate(self, validation_key: str, model: dict):
        """
        Validate based on a schema definition
        :param validation_key: path in swagger document to find schema
        :param model: model to validate
        :return: None if validation is successful, otherwise an exception will be raised
        """
        spec = reach(self.spec_source, validation_key)
        if not spec:
            raise Base422Exception('No validation available',
                                   reason='A spec was requested but not resolved from API decs')
        try:
            validate_object(self.spec, spec, model)
        except ValidationError as verr:
            raise Base422Exception('Validation Failed', str(verr))

    def validate_request(self, resource: str, operation: str, model: RequestValidationModel):
        """
        Validate a 'request-like' object
        :param resource: API resource pointer
        :param operation: operation on resource
        :param model: model to validate
        :return: None if validation is successful, otherwise an exception will be raised
        """

        spec = self.spec.resources[resource].operations[operation]
        if not spec:
            raise Base422Exception('No validation available',
                                   reason='A spec was requested but not resolved from API decs')
        try:
            request_data = unmarshal_request(model, spec)
        except (ValidationError, BaseException) as err:
            raise Base422Exception('Validation Failed', str(err))
        return request_data



"""
Module for supporting a hard delete on a pynamo item
"""
from typing import Any

from wavelength.dal.dao.commands.post_filter_pynamo_command import PostFilteredPynamoCommand


class PynamoDeleteCommand(PostFilteredPynamoCommand):
    """
    Performs a permanent delete on an an item
    """

    def _execute(self) -> Any:
        """
        Run the command
        :return:Union[List[PynamoPersistenceModel],PynamoPersistenceModel]
        """
        result = super()._execute()
        return result.delete()

    def __call__(self):
        return self._execute()



"""
Observable Data Model
"""
from copy import copy
from typing import List, Any, Dict, Optional, Type

from pynamodb.models import Model as PynamoModel
from schematics.types import StringType, IntType

from wavelength.dal.dal_config import DalConfig
from wavelength.dal.dao.models.observable_persistence_model import ObservablePersistenceModel
from wavelength.model.model_base import BaseModel


class BasePersistenceModel(ObservablePersistenceModel):
    """
    Self-persisting base model
    """
    created_at = IntType(default=0)
    modified_at = IntType(default=0)
    version = IntType(default=1)
    table_state = StringType(default=DalConfig.table_state_new)

    def __init__(self, source_model: dict, protected_fields=None, column_mapping: dict = None, **kwargs):

        super().__init__(source_model, protected_fields=protected_fields, **kwargs)
        self._column_mapping = column_mapping if column_mapping else {}
        self.system_fields = ['created_at', 'modified_at', 'version', 'table_state']

    def get_mapped_changes(self):
        """
        Converts the change set to represent underlying schema
        :return: dict
        """
        changes = self._get_changes()
        if not self._column_mapping:
            return changes

        return self._resolve_mapped_changes(changes)

    def _resolve_mapped_changes(self, source: Dict[str, Any]) -> Dict[str, Any]:
        result = {}  # type: Dict[str, Any]
        for key in source:
            result.update(**self._resolve_mapped_change(key, source))
        return result

    def _resolve_mapped_change(self, key: str, changes: Dict[str, Any]) -> Dict[str, Any]:
        mapped_key = self._resolve_mapped_key(key)
        return {mapped_key: changes[key]}

    def _resolve_mapped_key(self, key: str) -> str:
        mapping = self._column_mapping.get(key)
        if mapping:
            mapped_key = mapping() if callable(mapping) else mapping
            return mapped_key
        return key

    def build(self, source_model: Dict[str, Any]) -> 'BasePersistenceModel':
        """
        Creates an instance of this model
        :param source_model:
        :return:
        """
        result = copy(self)
        result.from_json(self._resolve_mapped_changes(source_model))
        return result

    def push(self, source_model: Dict[str, Any]) -> None:
        """
        In-place swap of attribute values based on source model
        :param source_model: dict
        :return: None
        """
        self.from_json(self._resolve_mapped_changes(source_model))


class PynamoPersistenceModel(BasePersistenceModel):
    """
    Base model for developing domain models with a DynamoDB back end
    """

    def __init__(self, source_model: Dict[str, Any], db_model: Type[PynamoModel], **kwargs):
        super().__init__(source_model, **kwargs)
        self.db_model = db_model
        self.db_hash_key = None
        self.db_range_key = None
        self.non_key_db_attributes: List = []
        self._discover_db_attributes()

    def get_key(self) -> List[str]:
        """
        Gets keys for keyed model
        :return: list
        """
        result: List[str] = []
        if self.db_hash_key:
            result.append(getattr(self, self._resolve_mapped_key(self.db_hash_key)))

        if self.db_range_key:
            result.append(getattr(self, self._resolve_mapped_key(self.db_range_key)))

        return result

    def get_creation_actions(self) -> Dict[str, Any]:
        """
        Builds a dict representing non-key attributes from the db_model
         mapped to non-None values from the model
        :return: dict
        """
        result: Dict[str, Any] = {}
        for key in self.non_key_db_attributes:
            action = self._get_creation_action(key)
            if action:
                result.update(**action)

        return result

    def _get_creation_action(self, key: str) -> Optional[Dict[str, Any]]:
        if hasattr(self, key):
            value = getattr(self, key)
            if value is not None and isinstance(value, BaseModel):
                return {key: value.to_dict()}
            if value is not None:
                return {key: value}
        return None

    def _discover_db_attributes(self) -> None:
        """
        Loop through the Pynamodb attributes member and discover available fields
        :return: None
        """
        for prop, val in getattr(self.db_model, '_attributes').items():

            if val.is_hash_key:
                self.db_hash_key = prop
            elif val.is_range_key:
                self.db_range_key = prop
            else:
                self.non_key_db_attributes.append(prop)



"""
Base level errors
"""
from abc import abstractmethod
from enum import Enum
from http import HTTPStatus

from wavelength.util.standard_response import get_standard_response


class ExceptionLogLevel(Enum):
    """
    Enum representing the expected log level of an Exception.
    """
    EXCEPTION = 1
    WARNING = 2
    DEBUG = 3
    INFO = 4
    CRITICAL = 5
    ERROR = 6


BASE_EXCEPTION_EVENT_TYPE = 'EXCEPTION'


class WAVELENGTHBaseException(Exception):
    """
    Base exception for serverless. Other exceptions inherit from this class in
    order to provide a standard error message response format. Subclasses
    must therefor implemtned the defined abstract method(s) defined herein.
    """

    def __init__(self, error="Not Available", reason="Not Available", code="0"):
        super(WAVELENGTHBaseException, self).__init__(error)
        self.reason = reason
        self.code = code
        self.exception_log_level = ExceptionLogLevel.EXCEPTION
        self.event_type = BASE_EXCEPTION_EVENT_TYPE

    def _get_response(self, status_code, context):
        """
        Builds an error response dict based on the supplied status code and AWS context object.
        :param status_code: int HTTP Status Code (https://www.w3.org/Protocols/rfc2616/rfc2616-sec10.html).
        :param context: Python Lambda Context Object.
        (https://docs.aws.amazon.com/lambda/latest/dg/python-context-object.html#python-context-object-props).
        :return: dict
        """
        result = {
            "message": str(self),
            "error": {
                "path": context.invoked_function_arn,
                "reason": self.reason,
                "requestId": context.aws_request_id,
                "code": self.code
            }
        }
        return get_standard_response(result, status_code)

    @abstractmethod
    def get_response(self, context):
        """
        abstract method to required by inheriting classes to override.
        Retrieves the response string for the exception.
        """
        raise NotImplementedError()


class Base401Exception(WAVELENGTHBaseException):
    """
    401 Unauthorized.
    """

    def get_response(self, context=None):
        return self._get_response(HTTPStatus.UNAUTHORIZED, context)


class Base404Exception(WAVELENGTHBaseException):
    """
    404 Not found (API route).
    """

    def __init__(self, error="Not Available", reason="Not Available", code="0", object_id=None):
        super(Base404Exception, self).__init__(error)
        self.reason = reason
        self.code = code
        self.object_id = object_id
        if self.object_id is not None:
            self.reason = f'Unable to find a model with ID: {self.object_id}'

    def get_response(self, context=None):
        return self._get_response(HTTPStatus.NOT_FOUND, context)


class Base403Exception(WAVELENGTHBaseException):
    """
    403 Forbidden.
    """

    def get_response(self, context=None):
        return self._get_response(HTTPStatus.FORBIDDEN, context)


class Base409Exception(WAVELENGTHBaseException):
    """
    409 Conflict.
    """

    def get_response(self, context=None):
        return self._get_response(HTTPStatus.CONFLICT, context)


class Base412Exception(WAVELENGTHBaseException):
    """
    412 Precondition Failed exception.
    """

    def get_response(self, context=None):
        return self._get_response(HTTPStatus.PRECONDITION_FAILED, context)


class Base415Exception(WAVELENGTHBaseException):
    """
    415 Invalid Media Type exception.
    """

    def get_response(self, context=None):
        return self._get_response(HTTPStatus.UNSUPPORTED_MEDIA_TYPE, context)


class Base422Exception(WAVELENGTHBaseException):
    """
    422 Input validation error.
    """

    def get_response(self, context=None):
        return self._get_response(HTTPStatus.UNPROCESSABLE_ENTITY, context)


class Base429Exception(WAVELENGTHBaseException):
    """
    429 Too many request exception.
    """

    def get_response(self, context=None):
        return self._get_response(HTTPStatus.TOO_MANY_REQUESTS, context)


class Base5xxException(WAVELENGTHBaseException):
    """
    5xx Server Errors.
    """

    def get_response(self, context=None):
        return self._get_response(HTTPStatus.INTERNAL_SERVER_ERROR, context)


class InvalidParametersException(Base422Exception):
    """
    Derived input validation exception, use when parameters are not valid
    """


class OutputConversionException(Base422Exception):
    """
    Derived input validation exception, use when parameters are not valid
    """


class InvalidEventInputException(Base422Exception):
    """
    Derived input validation exception, use when lambda input event is not valid
    """



# # TODO - enable this once used on a project
# """
# Module to support creating a lock like object
# """
# import time
#
# from pynamodb.exceptions import PutError
# from pynamodb.expressions.condition import And, Or
# from pynamodb.expressions.operand import Path
#
# from wavelength.config import CONFIG
# from wavelength.dal.dao.commands.post_filter_pynamo_command import PostFilteredPynamoCommand
# from wavelength.dal.dao.models.platform_persistence_model import PynamoPersistenceModel
# from wavelength.dal.pynamo_models.util import pynamo_write_handler
# from wavelength.errors.exceptions import Base404Exception, Base409Exception
# from wavelength.model.util import get_posix_timestamp
#
#
# class AcquireLockFailureException(Base409Exception):
#     """
#     409 Exception unable to acquire lock for resource.
#     """
#
#
# class PynamoAquireLockCommand(PostFilteredPynamoCommand):
#     """
#     This class encapsulates creating a lock type object
#     """
#
#     def __init__(self, parent_model: PynamoPersistenceModel):
#         super().__init__(parent_model, post_filter=filter_active_locks)
#         self._validate_db_model_schema()
#         self.lock = None
#
#     def _validate_db_model_schema(self):
#         return (hasattr(self._parent.db_model, 'ttl') and
#                 hasattr(self._parent.db_model, 'created_at') and
#                 hasattr(self._parent.db_model, 'kind'))
#
#     def _execute(self):
#         """
#         Execute - Attempt to create a lock model and return the result if successful,
#          otherwise return None if one exists and isn't expired
#         :return: None
#         """
#         self.lock = self._create_lock()
#         if self.lock:
#             return self._parent.build(self.lock.to_json())
#         return self._check_lock_exists()
#
#     def __call__(self):
#         return self._execute()
#
#     def _check_lock_exists(self):
#         try:
#             result = super()._execute()
#         except Base404Exception:
#             # No model was found in super
#             return False
#
#         if result:
#             raise AcquireLockFailureException('Failed to lock. A different process is running.')
#
#         return False
#
#     @pynamo_write_handler
#     def _create_lock(self):
#         """
#         Create record
#         :return:None
#         """
#         now = int(time.time())
#         ttl = (now + CONFIG.LOCK_THRESHOLD)
#         created_at = get_posix_timestamp()
#         keys = self._parent.get_key()
#         if len(keys) > 1:
#             raise ValueError('Lock command only supports single keyed models')
#         record = self._parent.db_model(keys[0], created_at=created_at, ttl=ttl, kind=self._parent.kind)
#         try:
#             record.save(Or(self._parent.db_model.hash_id.does_not_exist(),
#                            And(self._parent.db_model.hash_id == Path(self._parent.hash_id),
#                                self._parent.db_model.ttl <= now)))
#         except PutError:
#             return None
#
#         return record
#
#
# def filter_active_locks(record):
#     """
#     Return True if a lock was found but the TTL was expired, otherwise return False
#     """
#     if record and record.ttl:
#         now = int(time.time())
#         return record.ttl > now
#     return False



"""
Default config for DAL
"""
from wavelength.dal.dynamodb.util import BotoThrottleConfig


class DalConfig(BotoThrottleConfig):
    """
    DAL configuration settings
    """
    results_default_limit: int = 10
    # Table States:
    table_state_new = 'NEW'
    table_state_modified = 'MOD'
    table_state_deleted = 'DEL'
    table_state_invalid = 'ERR'



"""
Custom Pynamodb attribute to support taxonomy.
"""
from functools import singledispatch

from pynamodb.attributes import UnicodeAttribute

from wavelength.model.util import safe_split


@singledispatch
def serialize_tag(source) -> None:
    """
    Default serializer for tags (if not supported by registered conversion we make it None).
    :param source: obj
    :return: None
    """


@serialize_tag.register(str)
def serialize_tag_str(source: str) -> str:
    """
    Tag is already a str, we are just returning it unmodified.
    :param source:
    :return: str
    """
    return source


@serialize_tag.register(list)
def serialize_tag_list(source: list) -> str:
    """
    Tag is a list, break it down to format.
    :param source:
    :return: str
    """
    if not source:
        return None

    source.sort()
    return f'::{"::".join(map(str, source))}::'


class TagAttribute(UnicodeAttribute):
    """
    A custom model activity.
    """

    def serialize(self, value):
        """
        When value is a list and return tag formatted str.
        When value is a str return str.
        Otherwise (in other cases) return None.
        :param value:
        :return:str
        """
        if value:
            return super(TagAttribute, self).serialize(serialize_tag(value))

        return value

    def deserialize(self, value: str):
        if value:
            if '::' in value:
                return super(TagAttribute, self).deserialize(safe_split(value, '::'))
            return super(TagAttribute, self).deserialize(value)

        return value



"""
Module for abstracting commands on a pynamo item
"""
from operator import and_, or_
from typing import Any, List, Callable, Dict

import simplejson as json
from pynamodb.expressions.condition import Comparison
from pynamodb.expressions.operand import Path as PynamoAction
from schematics.types import StringType, ListType, BaseType

from wavelength.dal.dao.models.platform_persistence_model import PynamoPersistenceModel
from wavelength.errors.exceptions import Base422Exception
from wavelength.model.model_base import BaseModel

OPERATOR_LOOKUP = {
    'eq': '__eq__',
    'ne': '__ne__',
    'lt': '__lt__',
    'gt': '__gt__',
    'lte': '__le__',
    'gte': '__ge__',
    'begins_with': 'startswith',
    'between': 'between',
    'is_in': 'is_in',
    'contains': 'contains'
}


def always_pass(record):
    """
    Pass-through filter
    :param record:
    :return:
    """
    return record


class FilterModel(BaseModel):
    """
    Class to encapsulate a filter condition
    """
    model_attribute = StringType()
    operator = StringType()
    combinator = StringType(default='and', choices=['and', 'or'])
    parameters = ListType(BaseType())

    def __init__(self, model_attribute: str, operator: str, parameters: Any):
        """
        Build a filter operation from args
        :param model_attribute: str
        :param operator: str
        :param parameters: Any
        """
        super().__init__()
        self.model_attribute = model_attribute
        self.operator = operator
        self.parameters = parameters

    def __call__(self) -> dict:
        """
        Build and return DAL struct for creating filtered expressions
        primarily used for backwards-compatibility
        :return: dict
        """
        return {
            'key': self.model_attribute,
            'operator': self.operator,
            'combinator': self.combinator,
            'args': self.parameters
        }

    def __str__(self):
        return json.dumps(self())


class BasePynamoCommand:
    """
    Base Command Interface
    """

    def __init__(self, parent_model: PynamoPersistenceModel):
        """
        :param parent_model: PynamoPersistenceModel (domain model object)
        """
        self._parent = parent_model

    def _execute(self) -> Any:
        """
        Call the command and optionally return a result
        :return:
        """
        raise NotImplementedError('Must override this in a child command class')

    def _build_model_update_actions(self) -> Dict[str, PynamoAction]:
        """
        Takes a dict representing changes from the parent model and converts them to
        Pynamodb update actions
        :return: list[PynamoAction]
        :raise: Base422Exception if model doesn't contain requested attribute
        """
        result = {}
        changes = {k: v for k, v in self._parent.get_mapped_changes().items() if v is not None}
        keys = changes.keys()
        for key in keys:
            prop = getattr(self._parent.db_model, key, None)
            if prop is None:
                raise Base422Exception(f'Model has no property named "{key}""')
            val = changes[key]
            if isinstance(val, list):
                val.sort()
            result.update({key: prop.set(val)})
        return result

    def _build_conditionals(self, filters: List[FilterModel]) -> Comparison:
        """
        Parses a generic DSL (key, operator, args) and converts it to pynamodb Conditions(conditional)
        :param filters: List[FilterModel]
        :return: Comparison
        """
        conditions = None

        for filter_ in filters:
            combinator = or_ if filter_.combinator == 'or' else and_
            key = getattr(self._parent.db_model, filter_.model_attribute, None)
            self._validate_filter_key(filter_, key)
            condition_operator = getattr(key, OPERATOR_LOOKUP.get(filter_.operator, filter_.operator), None)
            self._validate_filter_operator(filter_, condition_operator)
            args = filter_.parameters
            result = self._build_conditional(condition_operator, args)
            conditions = combinator(conditions, result) if conditions is not None else result

        return conditions

    @classmethod
    def _validate_filter_key(cls, filter_model, key):
        if key:
            return
        msg = f'Model has no property named "{filter_model.model_attribute}"'
        raise Base422Exception(msg)

    @classmethod
    def _validate_filter_operator(cls, filter_model, operator):
        if operator:
            return
        msg = f'Attribute {filter_model.model_attribute} does not support the operator "{filter_model.operator}"'
        raise Base422Exception(msg)

    @classmethod
    def _build_conditional(cls, operator: Callable, args):
        if isinstance(args, list):
            return operator(*args)
        return operator(args)

    def __repr__(self):
        return f'{type(self).__name__}({self.__dict__})'

    def __str__(self):
        return json.dumps(self.__dict__)



# pylint: disable=invalid-name
# pylint: disable=protected-access
"""
Lightweight config value resolver
"""
import os

from typing import Union, Dict


class Config:
    """
    Class that resolves env vars as getters that can be given default values
    via class level members

    Env value will always override
    """

    LOCK_THRESHOLD = 0.1, float

    # Default values and constants:
    DEFAULT_LAMBDA_EXEC_MIN_TIME = 10000, int  # time in millis (10 second default)
    DEFAULT_REQUEST_TIMEOUT = 10, int
    MAX_NUMBER_OF_CLIENT_RETRIES = 5, int
    CLIENT_RETRY_WAIT_EXPONENTIAL_MULTIPLIER = 500, int
    CLIENT_RETRY_WAIT_EXPONENTIAL_MAX = 10000, int

    DATA_CACHE_TTL = .001, float

    def __init__(self):
        """
        creates default values from class members
        """
        instance_type = type(self)
        defaults_key = f'_{instance_type.__name__}__defaults'
        parsers_key = f'_{instance_type.__name__}__parsers'
        self.parsers = getattr(instance_type, parsers_key) if hasattr(instance_type, parsers_key) else {}
        self.defaults = getattr(instance_type, defaults_key) if hasattr(instance_type,
                                                                        defaults_key) else self.get_members()

        instance_type.__defaults = self.defaults
        instance_type.__parsers = self.parsers
        for item in self.defaults:
            if hasattr(instance_type, item):
                delattr(instance_type, item)

    def __getattr__(self, name) -> Union[str, int, float, bool]:
        """
        Attempt to resolve value from os.environ with fallback support for constants
        :param name: attribute namestr
        :return: Union[str, int, float, bool]
        """
        result = os.environ.get(name, self.defaults.get(name))
        if result:
            if type(self).__parsers.get(name):
                return type(self).__parsers[name](result)
        return result

    def parse_value(self, key, val) -> str:
        """
        Support supplied type converter, cache for later resolution
        :param val: if tuple, attempt to use second part of tuple to
        ensure to convert the value, defaults to str
        :return: Union[str, int, float, bool]
        """
        if isinstance(val, tuple) and callable(val[1]):
            self.parsers[key] = val[1]
            return str(val[0])

        return str(val)

    def get_members(self) -> Dict[str, str]:
        """
        Gets class level attributes
        :return: dict
        """
        return {k: self.parse_value(k, v) for k, v in vars(type(self)).items()
                if k and not isinstance(v, property) and not (callable(v) or k.startswith("__"))}

    @property
    def stage(self):
        """
        stage is an overwrite property.
        """
        return self.stageName


CONFIG = Config()



"""
Boto utils
"""
from botocore.exceptions import ClientError


def get_boto3_error(error):
    """
    must be an instance of botocore.exceptions ClientError.
    """
    aws = {}
    message = ''

    # boto3 response
    if hasattr(error, 'response'):
        meta = error.response.get('ResponseMetadata')
        if meta:
            aws['status'] = meta['HTTPStatusCode']
        error = error.response.get('Error')
        if error:
            aws['code'] = error['Code']
            aws['message'] = error['Message']

    if hasattr(error, 'operation_name'):
        aws['op'] = error.operation_name

    if hasattr(error, 'message') or hasattr(error, 'Message'):
        message = error.message
    elif aws.get('message') or aws.get('Message'):
        message = aws['message']

    return {'message': message, 'aws': aws}


def is_a_boto3_throttle_error(error):
    """
    error must be an instance of botocore.exceptions ClientError.
    Determines if the error is a type of throttling error from AWS
    """
    if isinstance(error, ClientError):
        error = get_boto3_error(error)
    elif isinstance(error, dict) is False:
        return False

    http_status_code = error.get('aws', {}).get('status')
    if http_status_code in (429,):
        # too many requests = 429
        return True

    # not all AWS services adhere to the http request status code of 429 for
    # too many requests, do an error code check
    # this is fairly crude, to really do this properly exhaustively, have to
    # look at all the service specific schema files under botocore/data/
    # this just captures what are the retry codes that default botocore/data/_retry.json
    # looks for before it was set to an empty json object at build time.
    error_type = error.get('aws', {}).get('code')
    return error_type in ('Throttling', 'ThrottlingException',
                          'ProvisionedThroughputExceededException', 'RequestLimitExceeded',
                          'RequestThrottled', 'LimitExceededException',
                          'BandwidthLimitExceeded', 'TooManyRequestsException')



"""
Base pynamo command for gathering a single record and filtering it on the client side
"""
from typing import Callable, Optional, List, Union

from pynamodb.exceptions import DoesNotExist
from pynamodb.models import Model as PynamoModel

from wavelength.dal.dal_config import DalConfig
from wavelength.dal.dao.commands.pynamo_command import BasePynamoCommand, always_pass
from wavelength.dal.dao.models.platform_persistence_model import PynamoPersistenceModel
from wavelength.dal.pynamo_models.util import pynamo_read_handler
from wavelength.errors.exceptions import Base404Exception


class PostFilteredPynamoCommand(BasePynamoCommand):
    """
    Base Pynamo Command with post-operation filtering
    """

    def __init__(self, parent_model: PynamoPersistenceModel, post_filter: Callable = always_pass):
        """

        :param parent_model: PynamoPersistenceModel (domain model object)
        :param post_filter: callable->bool function to run retrieved items against
        """
        super().__init__(parent_model)
        self._post_filter: Callable = post_filter

    def _execute(self) -> Union[List[PynamoPersistenceModel], PynamoPersistenceModel]:
        """
        Run the command
        :return:Union[List[PynamoPersistenceModel],PynamoPersistenceModel]
        """
        return self._filter(self._gather(self._get_key()))

    def _get_key(self) -> List[str]:
        return self._parent.get_key()

    @pynamo_read_handler
    def _gather(self, key=None) -> Optional[PynamoModel]:
        """
        Pulls the most recent model from DB based on the parent model's keys
        :return: PynamoModel
        """
        try:
            if not key:
                key = self._parent.get_key()
            item = self._parent.db_model.get(*key, consistent_read=True)
            return item
        except DoesNotExist as err:
            raise Base404Exception('Model does not exist', object_id=key) from err

    def _filter(self, item: PynamoModel) -> Optional[PynamoModel]:
        """
        Perform post-result filtering
        :param item: PynamoModel
        :return: all items that yield a 'truthy' result from the _post_filter
        """

        if self._post_filter(item):
            return item
        return None


def filter_deleted_items(record) -> bool:
    """
    filter_item is a method that handles the filtering of items. It
    raises an exception depending on the state of the object defined
    in the function implementation.
    @return: True
    :raise Base404Exception if model is already deleted
    """
    if hasattr(record, 'table_state') and record.table_state == DalConfig.table_state_deleted:
        raise Base404Exception(
            'Model does not exist', object_id=record.to_json())
    return True



"""
Logging utils
"""
import json
import logging

MAX_LOG_OUTPUT = 20000
MAX_LOG_DEBUG_OUTPUT = 200000
MAX_LOG_APPEND = ' ...'
FORMAT = '%(message)s'

ENV_LOG_VAR_KEY = 'LOG_LEVEL'
DEFAULT_LOGGING_NAME = 'INFO'
DEFAULT_LOGGING = logging.INFO
LOG_NAME_VS_VAL = {
    'CRITICAL': logging.CRITICAL,
    'ERROR': logging.ERROR,
    'WARNING': logging.WARNING,
    'INFO': logging.INFO,
    'DEBUG': logging.DEBUG}
DEBUG_LOGGING = logging.DEBUG
# Limiting the log output, max for an AWS cloudwatch event is 256K we are
# limiting log events to ~20K for non debug and ~200K for debug.

WHITELISTPATH = ['interim_desc', 'requestContext', 'authorizer', 'claims']
WHITELISTATTRS = ['token_use', 'sub', 'locale', 'iat', 'email_verified', 'auth_time', 'aud']


class FluentLogger():
    """
    Structured logger wrapper, for fluent
    """

    def __init__(self, log_class, event=None, msg=None, **kwargs):
        """
        Standard constructor
        """
        self._event = event
        self._msg = msg
        self._limit_output = kwargs.pop('limit_output')
        self._previous_args = dict()
        self.append_previous(kwargs)
        self._log_class = log_class

    def exception(self, error=None, **kwargs):
        """
        log exception to structure logger
        """
        self._log('exception', error if error else self._event,
                  fluent_logger=self, **kwargs)
        return self

    def critical(self, event=None, msg=None, **kwargs):
        """log critical to structure logger - wrapper to _log method"""
        return self._log("critical", event=event, msg=msg, **kwargs)

    def error(self, event=None, msg=None, **kwargs):
        """log error to structure logger - wrapper to _log method"""
        return self._log("error", event=event, msg=msg, **kwargs)

    def info(self, event=None, msg=None, **kwargs):
        """log info to structure logger - wrapper to _log method"""
        return self._log("info", event=event, msg=msg, **kwargs)

    def debug(self, event=None, msg=None, **kwargs):
        """log debug to structure logger - wrapper to _log method"""
        return self._log("debug", event=event, msg=msg, **kwargs)

    def warn(self, event=None, msg=None, **kwargs):
        """log warn to structure logger - wrapper to _log method"""
        return self._log("warn", event=event, msg=msg, **kwargs)

    def append_previous(self, previous_args):
        """
        Helper to update previous args
        """
        if previous_args:
            self._previous_args.update(previous_args)
            if 'exc_info' in self._previous_args:
                self._previous_args.pop('exc_info')
        return self._previous_args

    def _log(self, cmd, event=None, msg=None, **kwargs):
        """
        log critical to structure logger
        """
        limit_output = kwargs.get('limit_output', True)
        kwargs['fluent_logger'] = self

        self.append_previous(kwargs)
        limit_output = self._limit_output if not limit_output else limit_output
        getattr(self._log_class, cmd)(
            event=event if event else self._event,
            msg=msg if msg else self._msg,
            **self._previous_args)
        return self


class BufferStreamHandler(logging.StreamHandler):
    """
        BufferStreamHandler is a subclass to logging.StreamHandler that is
        used to buffer messages to then be flushed at once.
    """

    LOGGABLE_RECORDS = set([logging.ERROR, logging.CRITICAL])

    def __init__(self, stream=None):
        """
        Initialize the handler.

        If stream is not specified, sys.stderr is used.
        """
        super().__init__(stream)
        self.message_buffer = {'items': []}

    def add(self, key, value):
        """
        Appends and item by namespace it to a different key.
        """
        self.message_buffer[key] = value

    def emit(self, record):
        """
        Emit a record.
        """

        try:
            msg = self.format(record)
            try:
                dmsg = json.loads(msg)
                self.message_buffer['items'].append(dmsg)
            except json.JSONDecodeError:
                self.message_buffer['items'].append(msg)
            finally:
                if record.levelno in BufferStreamHandler.LOGGABLE_RECORDS:
                    self.write_to_stream(msg)
        # pylint: disable=W0703
        except Exception:
            self.handleError(record)

    def write_to_stream(self, msg):  # msg is a string
        """
        writes message to a buffer.
        """
        stream = self.stream
        stream.write(msg)
        stream.write(self.terminator)
        self.flush()

    def flush_buffer_to_stream(self):
        """
        Flushes buffer by writing to stream.
        """
        if self.message_buffer['items']:
            jmsg = json.dumps(self.message_buffer)
            self.write_to_stream(jmsg)
            self.message_buffer = {'items': []}


def fit_log_msg(msg, limit_output, max_log_msg_length):
    """
    Trims the log based on maximum length declared for debug or regular messages
    """
    if not limit_output:
        return msg
    msg_str = str(msg)
    return (msg_str[:max_log_msg_length] +
            MAX_LOG_APPEND) if len(msg_str) > max_log_msg_length else msg



"""
Finds delta on two dicts for things like change detection
"""


class DictDiffer:
    """
    Calculate the difference between two dictionaries as:
    (1) items added
    (2) items removed
    (3) keys same in both but changed values
    (4) keys same in both and unchanged values
    """

    def __init__(self, new_dict, past_dict):
        """
        Constructor pass two dicts current and the original
        """

        self._new_dict, self._past_dict = new_dict, past_dict
        self._set_new, self._set_past = set(new_dict.keys()), set(past_dict.keys())
        self._intersect = self._set_new.intersection(self._set_past)

    def added(self):
        """
        Get keys of added dict items to new dictionary
        """
        return self._set_new - self._intersect

    def removed(self):
        """
        Get removed keys from old dict to new
        """
        return self._set_past - self._intersect

    def changed(self):
        """
        Get keys of changed dict values of the same key
        """
        return set(o for o in self._intersect if self._past_dict[o] != self._new_dict[o])

    def unchanged(self):
        """
        Get unchanged values of the same key
        """
        return set(o for o in self._intersect if self._past_dict[o] == self._new_dict[o])



"""
Common utils for extracting info from AWS data structures
"""
import json
import os

from schleppy import reach

from wavelength.errors.exceptions import Base422Exception
from wavelength.util import standard_response

HEADER_PROFILE_INDEX_KEY = 'PROFILE_IDX'
EIDS_CUSTOM_CLAIM = "custom:eids"
PROFILES_CUSTOM_CLAIM = "custom:profiles"


def get_stage(context=None):
    """
    Retrieves the stage based on a few possible input(s)
    is stage declared in the event.
    is stage prefixed in the function name in the context.
    is stage prefixed in the function name as an env variable.
    """
    if context is not None:
        tokens = context.function_name.split('-')
    else:
        tokens = os.environ.get('AWS_LAMBDA_FUNCTION_NAME', '').split('-')
    return tokens[-2] if (len(tokens) > 1) else None


def convert_body(body):
    """
    Attempts to convert the body of an event to a json object.
    This is a general function not necessarily just for events.
    In the event it is unable to convert to json the original input is returned.
    """
    if body and isinstance(body, str):
        try:
            return json.loads(body)
        except Exception:  # pylint: disable=broad-except
            pass
    return body


def convert_response(response):
    """
    Attempts to converts the body of an event from a str to json object.
    """
    if response and isinstance(response, dict):
        body = response.get('body')
        if body is not None and not isinstance(body, str):
            try:
                response['body'] = json.dumps(body)
            except Exception:  # pylint: disable=broad-except
                pass
    return response


def get_profile_id(event):
    """
    Gets the profile id from an event containing the authorization markup from a AWS authorizer
    Default behavior,
    if the OF:PROFILE_IDX header is not set it defaults to the first profile
    If the profiles field is not set in the authorization code it defaults to the first profile.

    return profile id or None, if the event is malformed, missing 'sub' in a claim

    Raises Base422Exception if profiles are defined as a claim and the index in the header is out of range
    """
    account_id = get_account_id(event)
    if not account_id:
        return None

    index = 1
    headers = get_headers(event)
    if headers:
        index = int(headers.get(HEADER_PROFILE_INDEX_KEY, '1'))
        if (index > 1) and (index > len(get_profiles(event))):
            raise Base422Exception('Invalid Profile Index Header, index value outside of profiles list length')
    return account_id + f'-{index}'


def get_headers(event):
    """Gets the headers from the standard lambda event."""

    return event.get('headers', {})


def get_authorization_header(event):
    """Gets the authorization header from the standard lambda event, returns None if not present."""

    return reach(event, "headers.Authorization", options={'default': None})


def get_claims(event):
    """Attempts to get the dictionary of claims from within standard lambda event."""

    return reach(event, 'requestContext.authorizer.claims', options={'default': {}})


def get_account_id(event):
    """Gets the account id from a standard lambda event."""

    return reach(event, 'requestContext.authorizer.claims.sub', options={'default': None})


def get_entitlements(event):
    """Gets the list of entitlement from within standard lambda event."""

    claims = get_claims(event)
    if claims:
        return convert_body(claims.get(EIDS_CUSTOM_CLAIM, {'eids': []})).get('eids')
    return {}


def get_profiles(event):
    """Gets the list of profiles from a claim within standard lambda event."""
    claims = get_claims(event)
    if claims:
        return convert_body(claims.get(PROFILES_CUSTOM_CLAIM, {'profiles': []})).get('profiles')
    return []


def get_standard_response(body, status_code=200, headers=None, add_cors=True):
    """
    Creates the standard response for a lambda return.
    params
        body: a string or object to return, do not covert using json dumps before,
               this will be handled in the log lambda wrapper to enahnce return
               logging by keeping the response as an object until it is returned.
        status_code: standard HTTP status code result, use the standard exception to handle non 200 and 300 results.
                     Do not handle 400 or 500 errors here. Raise and standard 400 or 500 exception will be handled by
                     the log lambda decorator
        headers: Any additional non CORS header, for CORS header use the add_cors parameter (true by default_
        add_cors: Adds cors headers, default is true (always)
    """
    return standard_response.get_standard_response(body, status_code, headers, add_cors)  # Backward compatible



"""
Utils for supporting validating API Gateway events
"""
from functools import wraps

import simplejson
from bravado_core.exception import SwaggerMappingError

from wavelength.errors.exceptions import Base422Exception


def validate_event(model, validation_exceptions=(SwaggerMappingError, Exception)):
    """
    Parses a Lambda Event, validates it and adds it to event[model]
    :param model: Callable(<address>,<event>) that the event is passed to for validation and transform.
    :param validation_key: Key address in swagger spec to find validation schema
    :param validation_exceptions: Exceptions to listen for to call 422 exceptions
    :return:
    """

    def package(func):
        """
        Standard decorator
        """

        @wraps(func)
        def wrapper(*args, **kwargs):
            """
            Validation wrapper
            """
            event = args[0]
            try:
                event['model'] = model(event)
            except validation_exceptions as vex:
                if any(isinstance(vex, ex) for ex in validation_exceptions):
                    if hasattr(vex, "reason"):
                        reason = vex.reason
                    else:
                        reason = str(vex)

                    raise Base422Exception('Validation Failed', reason=reason)
            try:
                return func(*args, **kwargs)
            finally:
                del event['model']

        return wrapper

    return package


class RequestValidationModel:
    """
    Base Validation model
    """

    def __init__(self, event_data):
        self._data = event_data

    @property
    def profile_id(self) -> str:
        """
        Profile Id
        :return: dict
        """
        # TODO - update to pull from Matt's JWT
        return self.headers.get('profile_id')

    @property
    def headers(self) -> dict:
        """
        Header Parameters
        :return: dict
        """
        return _sanitize_event(self._data.get('headers', {}))

    @property
    def query(self) -> dict:
        """
        Querystring Parameters
        :return: dict
        """
        return _sanitize_event(self._data.get('queryStringParameters', {}))

    @property
    def path(self) -> dict:
        """
        Path Parameters
        :return: dict
        """
        return _sanitize_event(self._data.get('pathParameters', {}))

    @property
    def form(self) -> dict:
        """
        Form Body
        :return: dict
        """
        return _sanitize_event(self._data.get('body', {}))

    @property
    def body(self) -> dict:
        """
        Request Body
        :return: dict
        """
        return _sanitize_event(self._data.get('body', {}))

    def json(self) -> str:
        """
        Request Body as JSON
        :return: dict
        """

        result = self._data.get('body', {})
        if isinstance(result, str):
            try:
                result = simplejson.loads(result)
            except(simplejson.JSONDecodeError, BaseException):
                result = {}

        return result


def _sanitize_event(val):
    """
    Return empty dictionary object here if val is null because Bravado doesn't null check
    """
    if val is None:
        val = {}
    return val


class NoAuthRequestValidationModel(RequestValidationModel):
    """
    Base Validation model with no Auth
    """

    @property
    def profile_id(self) -> str:
        """
        Profile Id
        :return: dict
        """
        return self._data.get('pathParameters', {}).get('profile-id')



"""
Observable Data Model
"""
from typing import List, Any, Dict

from wavelength.model.model_base import BaseModel


class ObservablePersistenceModel(BaseModel):
    """
    Self-persisting base model
    """

    def __init__(self, source_model: dict = None, protected_fields=None,
                 forced_change_fields: List[str] = None, **kwargs):
        """
        Creates an instance of a change tracking persistence model
        :param source_model: dict to hydrate model from
        :param dao: DaoBase to execute commands against
        :param protected_fields: fields to prevent marking as 'dirty'
        """
        super().__init__(raw_data=source_model, **kwargs)
        self._protected_fields = protected_fields if protected_fields else []
        self._changes = {}  # type: Dict[str, Any]
        self._forced_change_fields = forced_change_fields if forced_change_fields else []

    @property
    def is_dirty(self) -> bool:
        """
        If non-protected changes have been set on this model wil return True
        :return: bool
        """
        return bool(self._changes)

    def _get_changes(self) -> Dict[str, Any]:
        forced_changes = {}  # type: Dict[str, Any]
        for field in self._forced_change_fields:
            forced_changes.update(**{field: getattr(self, field)})
        return {**self._changes, **forced_changes}

    def flush_changes(self) -> 'ObservablePersistenceModel':
        """
        Clears the change set and un flags the model as being dirty
        :return: self
        """
        self._changes.clear()
        return self

    def __setattr__(self, key, value):
        # _fields is attached to the class during the schematics model instance creation process
        # pylint: disable=no-member
        if key in self._fields:
            if key in self._protected_fields:
                return
            self._changes.update(**{key: value})
        super().__setattr__(key, value)

    def override_protected_field(self, field_name, value) -> None:
        """
        Temporarily removes protected field to set value internally
        :param field_name: attribute to set
        :param value: new value
        :return: None
        """

        if not hasattr(self, field_name):
            raise AttributeError(f'No field named {field_name} on current model')
        if not isinstance(self, ObservablePersistenceModel):
            raise TypeError('_override_protected_field method should only be called from a model')
        if field_name not in self._protected_fields:
            setattr(self, field_name, value)
            return
        self._protected_fields.remove(field_name)
        setattr(self, field_name, value)
        self._protected_fields.append(field_name)



"""
Module to support updating a single record
"""
from typing import List, Callable, Optional, Iterator

from cachetools import cached, TTLCache
from pynamodb.models import Model as PynamoModel

from wavelength.config import CONFIG
from wavelength.dal.dal_config import DalConfig
from wavelength.dal.dao.commands.post_filter_pynamo_command import PostFilteredPynamoCommand, filter_deleted_items
from wavelength.dal.dao.commands.pynamo_command import FilterModel
from wavelength.dal.dao.commands.pynamo_query_args import PynamoQueryArguments
from wavelength.dal.dao.models.platform_persistence_model import PynamoPersistenceModel
from wavelength.dal.dynamodb.util import decode_start_token
from wavelength.dal.pynamo_models.util import pynamo_read_handler


class PynamoQueryResult:
    """
    Class to take a raw query result (ResultIterator) and form a typed response for encoded
    start token and domain model
    """

    def __init__(self, query_result: List[Optional[PynamoModel]],
                 parent_model: PynamoPersistenceModel, start_token: str = None) -> None:
        """
        Converts a result set from a pynamoDB query and converts it a collection of domain models
        :param query_result: List[PynamoModel]
        :param parent_model: PynamoPersistenceModel
        """
        self._start_token = start_token

        self._items = [parent_model.build(item.to_json()) for item in query_result]

    @property
    def start_token(self) -> Optional[str]:
        """
        Provides encoded start_token
        :return: str
        """
        return self._start_token

    @property
    def items(self) -> List[PynamoPersistenceModel]:
        """
        List of converted domain items
        :return: List[PynamoPersistenceModel]
        """
        return self._items


class PynamoQueryCommand(PostFilteredPynamoCommand):
    """
    This class encapsulates the updating of a versioned record
    """

    def __init__(self, parent_model: PynamoPersistenceModel,
                 pre_filters: List[FilterModel] = None,
                 post_filter: Callable = filter_deleted_items):
        """
        :param parent_model: PynamoPersistenceModel
        :param pre_filters: pre_filters: List[FilterModel])
        """
        super().__init__(parent_model, post_filter)
        self._next_token = None
        # using `is not None` here so we can clear out the pre_filters with an empty list
        self._pre_filters = pre_filters if pre_filters is not None else [
            FilterModel('table_state', 'is_in', [DalConfig.table_state_new, DalConfig.table_state_modified])]

    @cached(TTLCache(maxsize=32, ttl=CONFIG.DATA_CACHE_TTL))
    def __call__(self, query: PynamoQueryArguments) -> PynamoQueryResult:
        """
        Builds the query object for subsequent execution and returns the result
        :param query: str
        :return: PynamoQueryCommand
        """
        query.append_filters(self._pre_filters)
        return self._execute_query(query)

    def _execute_query(self, query: PynamoQueryArguments) -> PynamoQueryResult:
        """
        Execute command interface
        :return: list[PynamoPersistenceModel]
        """
        query_result = self._gather_query(query)
        post_filter: Callable = self._post_filter

        if query.post_filter:
            self._post_filter = query.post_filter

        result = PynamoQueryResult([self._filter(item) for item in query_result.items if item], self._parent,
                                   self._next_token)

        self._post_filter = post_filter
        self._next_token = None
        return result

    @pynamo_read_handler
    def _gather_query(self, query: PynamoQueryArguments) -> Iterator[PynamoModel]:
        """
        Run query and cast results to domain objects
        :return: QueryResult
        """

        return self._parent.db_model.query(query.hash_key,
                                           index_name=query.index_name,
                                           range_key_condition=self._build_conditionals(query.range_key_filter),
                                           filter_condition=self._build_conditionals(query.filters),
                                           limit=query.limit,
                                           scan_index_forward=query.scan_index_forward,
                                           consistent_read=query.consistent_read,
                                           last_evaluated_key=decode_start_token(query.start_token))



"""
Observable Data Model with CRUD OPs and DynamoDB back end
"""

from typing import Type

from pynamodb.models import Model as PynamoModel

from wavelength.dal.dao.commands.create_pynamo_command import PynamoCreateCommand
from wavelength.dal.dao.commands.get_by_key_pynamo_command import PynamoGetByIdCommand
from wavelength.dal.dao.commands.query_pynamo_command import PynamoQueryCommand
from wavelength.dal.dao.commands.soft_delete_pynamo_command import PynamoSoftDeleteCommand
from wavelength.dal.dao.commands.update_pynamo_command import PynamoUpdateCommand
from wavelength.dal.dao.models.platform_persistence_model import PynamoPersistenceModel


class PynamoCrudModel(PynamoPersistenceModel):
    """
    Base model for developing domain models with a DynamoDB back end
    """

    def __init__(self, source_model: dict, db_model: Type[PynamoModel], **kwargs):
        super().__init__(source_model, db_model, **kwargs)

        self.query = PynamoQueryCommand(self)
        self.create = PynamoCreateCommand(self)
        self.read = PynamoGetByIdCommand(self)
        self.update = PynamoUpdateCommand(self)
        self.delete = PynamoSoftDeleteCommand(self)

    def __repr__(self):
        return f'{type(self).__name__}({self.to_json()})'



"""
Query args for gathering a single record and filtering it on the client side
"""
from typing import List, Any, Callable

from wavelength.dal.dao.commands.post_filter_pynamo_command import filter_deleted_items
from wavelength.dal.dao.commands.pynamo_command import FilterModel
from wavelength.dal.dynamodb.util import decode_start_token


class PynamoQueryArguments:
    """
    Class to represent *args and **kwargs for a query method
    """

    def __init__(self, hash_key: str, **kwargs):
        """
        Represents input to the query command
        :param hash_key: str
        :param range_key_filter: str
        :param filters: List[FilterModel]
        :param scan_index_forward: bool
        :param limit: int
        :param start_token: str
        :param consistent_read: str
        """
        self.hash_key = hash_key
        self.index_name = kwargs.get('index_name')
        self.range_key_filter = kwargs.get('range_key_filter', [])
        self.filters = kwargs.get('filters', [])
        self.scan_index_forward = kwargs.get('scan_index_forward', False)
        self.limit = kwargs.get('limit', 30)
        start_token = kwargs.get('start_token')
        self.start_token = decode_start_token(start_token) if start_token else None
        self._consistent_read = kwargs.get('consistent_read', False)
        self._frozen = kwargs.get('frozen', False)
        self._post_filter: Callable = kwargs.get('post_filter', filter_deleted_items)

    @property
    def consistent_read(self) -> bool:
        """
        Property to enable / disable consistent reads bypassing the dal cache.
        """
        return self._consistent_read

    @consistent_read.setter
    def consistent_read(self, value: bool) -> None:
        """
        Setter for consistent_read property to enable / disable consistent reads bypassing the dal cache.
        """
        self._consistent_read = value

    @property
    def post_filter(self) -> Callable:
        """
        Query result filter function
        :return:
        """
        return self._post_filter

    def append_filters(self, filters: List[FilterModel]) -> None:
        """
        Adds filters to the list of filters
        :param filters: List[FilterModel]
        :return: None
        """
        if not self._frozen and filters:
            self.filters.extend(filters)

    def __hash__(self) -> Any:
        filters = '::'.join([str(filter_) for filter_ in self.range_key_filter + self.filters])

        return hash((self.hash_key, self.start_token or '', self.limit, self.scan_index_forward, filters))



"""
Custom Pynamodb attribute for epoch datetimes.
"""
from decimal import Decimal
from functools import singledispatch

import arrow

from pynamodb.attributes import NumberAttribute

from wavelength.model.util import get_posix_timestamp


@singledispatch
def serialize_epoch(source) -> None:
    """
    Default serializer for epochs (if not supported by registered conversion we make it None).
    :param source: obj
    :return: None
    """
    pass  # pylint: disable=W0107


@serialize_epoch.register(str)
def serialize_epoch_str(source: str) -> int:
    """
    Only date-time formmated strings are supported, parse it down to an int.
    :param source:
    :return: int
    """
    try:
        return get_posix_timestamp(arrow.get(source))
    except Exception:
        raise ValueError('Unable to parse string as a datetime')


@serialize_epoch.register(int)
def serialize_epoch_int(source: int) -> int:
    """
    Epoch is an int, pass it through
    :param source:
    :return: int
    """

    if isinstance(source, bool):
        raise ValueError('Boolean not supported for Epoch')

    return source


@serialize_epoch.register(float)
@serialize_epoch.register(Decimal)
def serialize_epoch_numeric(source) -> int:
    """
    Epoch is numeric but not an int, round it
    :param source:
    :return: int
    """

    return round(source)


class EpochAttribute(NumberAttribute):
    """
    A custom model activity.
    """

    def serialize(self, value):
        """
        When value is a list and return epoch formatted str.
        When value is a str return str.
        Otherwise (in other cases) return None.
        :param value:
        :return:str
        """
        if value:
            return super(EpochAttribute, self).serialize(serialize_epoch(value))

        return value



"""
Model utility functions
"""
from datetime import timezone, datetime, timedelta
from typing import Callable, Union


def get_posix_timestamp(datetime_value=None):
    """
    :param datetime_value: datetime
    :return: milliseconds from epoch
    """
    if datetime_value is None:
        datetime_value = datetime.utcnow()
    datetime_value = datetime_value.replace(tzinfo=timezone.utc)
    epoch = datetime(1970, 1, 1, tzinfo=timezone.utc)  # use POSIX epoch
    return round((datetime_value - epoch) / timedelta(microseconds=1))


def get_posix_date(epoch_seconds_value):
    """
    :param epoch_seconds_value: epoch milliseconds
    :return: datetime
    """
    try:
        return (datetime(1970, 1, 1, tzinfo=timezone.utc) +
                timedelta(microseconds=epoch_seconds_value))
    except (ValueError, TypeError):
        return None


def get_range_key(key, date=None, reverse=False) -> str:
    """
    :return: str
    """
    if reverse:
        template = '{1}::{0}'
    else:
        template = '{0}::{1}'

    if date is None:
        date = datetime.utcnow()

    if isinstance(date, datetime):
        return template.format(get_posix_timestamp(date), key)
    return template.format(key, date)


def safe_split(source: str, delimiter=',', converter: Callable = lambda x: x):
    """
    Splits a string via the delimiter, throws out any empties and strips the whitespace from each item
    :param source: str
    :param delimiter: str
    :param converter: Callable
    :return: list
    """
    if not source:
        return []
    return list(filter(None, [converter(item.strip()) for item in source.split(delimiter)]))


def safe_bool(source: Union[bool, str]):
    """
    Returns a boolean value from semantic string
    :param source: str
    :return: bool
    """
    if isinstance(source, str):
        return source.strip().lower() in ['true', 'yes', '1', 'ok']
    return source



"""
Base logger
"""
import logging
import os
import sys

from structlog._config import wrap_logger
from structlog.processors import (TimeStamper,
                                  StackInfoRenderer,
                                  format_exc_info,
                                  JSONRenderer)
from structlog.stdlib import add_logger_name, add_log_level

from wavelength.aws.aws_args import get_stage
from wavelength.errors.exceptions import ExceptionLogLevel, WAVELENGTHBaseException
from wavelength.logging.filter import NestedFilter, whitelist
from wavelength.logging.logging_util import (LOG_NAME_VS_VAL,
                                                      ENV_LOG_VAR_KEY, DEFAULT_LOGGING_NAME,
                                                      DEFAULT_LOGGING, MAX_LOG_DEBUG_OUTPUT,
                                                      DEBUG_LOGGING, MAX_LOG_OUTPUT,
                                                      FORMAT,
                                                      WHITELISTATTRS, WHITELISTPATH,
                                                      fit_log_msg)


class LogLambdaBase:
    """
    This is a decorator that will record the invoking of a lambda handler
    and its input event.

    Also, it will wrap the handler to record the exceptions in a structured
    format for ingestion into the log monitor es cluster.

    the decorator should be initialized with the logger returned
    from _get_structured_logger()
    """
    _log_handler = None
    _structured_logger = None
    _do_convert_body = None
    _max_log_msg_length = MAX_LOG_DEBUG_OUTPUT
    _log_level = logging.INFO

    # pylint: disable=too-many-arguments

    def __init__(self, service, logger_name=None, event=None, context=None, do_convert_body=True):
        """
        we will create the structured logger internally
        """
        LogLambdaBase._do_convert_body = do_convert_body
        self.service_tag = service

        # TODO should we use string values to represent log levels via env
        # variables?
        LogLambdaBase._log_level = LOG_NAME_VS_VAL.get(os.getenv(ENV_LOG_VAR_KEY,
                                                                 os.getenv(ENV_LOG_VAR_KEY.lower(),
                                                                           DEFAULT_LOGGING_NAME)),
                                                       DEFAULT_LOGGING)
        LogLambdaBase._max_log_msg_length = MAX_LOG_DEBUG_OUTPUT if self._log_level == DEBUG_LOGGING else MAX_LOG_OUTPUT

        if event is not None and context is not None:
            # Special constructor, if we want to do bindings before calling
            # the method A fully qualified call may be used construct logger
            # instead of implicitly constructor it during the call method
            context.structured_logger = self._set_structured_logger(
                context, logger_name=logger_name, service_tag=self.service_tag)

    @staticmethod
    def _set_structured_logger(context, logger_name=None, service_tag=None):
        """
            use this function to get a structured logger and can use
            this to pass into the LogLambda decorator

            context - aws lambda context
            service - name of the service that this lambda belongs to
            logger_name - logger's name
            log_level - one of the levels in logging module
        """
        if LogLambdaBase._structured_logger:
            return LogLambdaBase._structured_logger

        stage_tag = get_stage(context)

        if logger_name:
            logger = logging.getLogger(str(logger_name))
        else:
            logger = logging.getLogger()

        # Python logger in AWS Lambda has a preset format.
        # To change the format of
        # the logging statement, remove the logging handler
        # and add a new handler with the required format
        for handler in logger.handlers:
            logger.removeHandler(handler)

        LogLambdaBase._log_handler = LogLambdaBase._get_handler()

        LogLambdaBase._log_handler.setFormatter(logging.Formatter(FORMAT))
        logger.addHandler(LogLambdaBase._log_handler)
        logger.setLevel(LogLambdaBase._log_level)
        logger.propagate = False

        wlogger = wrap_logger(
            logger,
            processors=[
                LogLambdaBase._filter_pii_info,
                add_logger_name,
                add_log_level,
                TimeStamper(fmt="iso"),
                StackInfoRenderer(),
                format_exc_info,
                JSONRenderer(separators=(',', ':'), sort_keys=True)])

        inferred_lambda_tag = context.function_name
        if stage_tag is not None:
            inferred_lambda_tag = inferred_lambda_tag.replace('{0}_'.format(stage_tag), '', 1)

        LogLambdaBase._structured_logger = wlogger.bind(
            aws_lambda_name=context.function_name,
            aws_lambda_request_id=context.aws_request_id,
            internal_service_tag=service_tag,
            inferred_stage_tag=stage_tag,
            inferred_lambda_tag=inferred_lambda_tag
        )
        return LogLambdaBase._structured_logger

    @staticmethod
    def bind(**value):
        """
        add binding arguments to structured logger
        """
        if LogLambdaBase._structured_logger:
            LogLambdaBase._structured_logger = LogLambdaBase._structured_logger.bind(**value)

    @staticmethod
    def log_wavelength_exception(wavelength_exception: WAVELENGTHBaseException, fluent_logger=None, **kwargs):
        """
        log WAVELENGTHBaseException to the configured log level.
        """
        log_func = LogLambdaBase._get_wavelength_exception_log_function(wavelength_exception.exception_log_level)

        return log_func(wavelength_exception.event_type,
                        str(wavelength_exception),
                        exception_type=type(wavelength_exception).__name__,
                        exc_info=True,
                        fluent_logger=fluent_logger,
                        **kwargs)

    @staticmethod
    def _get_wavelength_exception_log_function(exception_log_level: ExceptionLogLevel):
        """
        Return the log function appropriate for the ExceptionLogLevel
        """
        _wavelength_exception_log_level_funcs = {
            # EXCEPTION type is deliberately logged as error
            ExceptionLogLevel.EXCEPTION: LogLambdaBase.error,
            ExceptionLogLevel.WARNING: LogLambdaBase.warn,
            ExceptionLogLevel.DEBUG: LogLambdaBase.debug,
            ExceptionLogLevel.INFO: LogLambdaBase.info,
            ExceptionLogLevel.CRITICAL: LogLambdaBase.critical,
            ExceptionLogLevel.ERROR: LogLambdaBase.error
        }

        return _wavelength_exception_log_level_funcs.get(exception_log_level, LogLambdaBase.error)

    @staticmethod
    def exception(error, fluent_logger=None, **kwargs):
        """
        log exception to structure logger
        """
        return LogLambdaBase.error('EXCEPTION',
                                   str(error),
                                   exception_type=type(error).__name__,
                                   exc_info=True,
                                   fluent_logger=fluent_logger,
                                   **kwargs)

    @staticmethod
    def critical(event, msg, limit_output=True, fluent_logger=None, **kwargs):
        """log critical to structure logger - wrapper to _log classmethod"""
        return LogLambdaBase.process_log("critical", event, msg,
                                         limit_output=limit_output,
                                         fluent_logger=fluent_logger, **kwargs)

    @staticmethod
    def error(event, msg, limit_output=True, fluent_logger=None, **kwargs):
        """log error to structure logger - wrapper to _log classmethod"""
        return LogLambdaBase.process_log("error", event, msg,
                                         limit_output=limit_output,
                                         fluent_logger=fluent_logger, **kwargs)

    @staticmethod
    def info(event, msg, limit_output=True, fluent_logger=None, **kwargs):
        """log info to structure logger - wrapper to _log classmethod"""
        return LogLambdaBase.process_log("info", event, msg,
                                         limit_output=limit_output,
                                         fluent_logger=fluent_logger, **kwargs)

    @staticmethod
    def debug(event, msg, limit_output=True, fluent_logger=None, **kwargs):
        """log debug to structure logger - wrapper to _log classmethod"""
        return LogLambdaBase.process_log("debug", event, msg,
                                         limit_output=limit_output,
                                         fluent_logger=fluent_logger, **kwargs)

    @staticmethod
    def warn(event, msg, limit_output=True, fluent_logger=None, **kwargs):
        """log warn to structure logger - wrapper to _log classmethod"""
        return LogLambdaBase.process_log("warning", event, msg,
                                         limit_output=limit_output,
                                         fluent_logger=fluent_logger, **kwargs)

    @staticmethod
    def _get_structured_logger():
        return LogLambdaBase._structured_logger

    @staticmethod
    def _clear_structured_logger():
        LogLambdaBase._structured_logger = None

    @staticmethod
    def _get_handler():
        return logging.StreamHandler(sys.stdout)

    @staticmethod
    def _filter_pii_info(_logger, _method_name, event_dict):
        whitelist_filter = NestedFilter(WHITELISTPATH, WHITELISTATTRS)
        return whitelist(event_dict, whitelist_filter)

    @staticmethod
    def _flush_buffer(**kwargs):
        LogLambdaBase._log_handler.flush_buffer_to_stream()

    @staticmethod
    def process_log(cmd, event, msg, **kwargs):
        """
        Standard base logger
        """
        if 'fluent_logger' in kwargs:
            kwargs.pop('fluent_logger')
        limit_output = kwargs.pop('limit_output', True)
        if LogLambdaBase._get_structured_logger():
            message = fit_log_msg(event, limit_output, LogLambdaBase._max_log_msg_length)
            desc = fit_log_msg(msg, limit_output, LogLambdaBase._max_log_msg_length)
            getattr(LogLambdaBase._get_structured_logger(), cmd)(message, interim_desc=desc, **kwargs)
            # a bit of meta programming here in order to remove duplication
            # in that each logging method basically has the same params with a
            # different message name. Using getattr is just as good.



"""
Utilities to check if inputs can be converted to various outputs
"""
import json


def is_int_like(obj):
    """
     provided an object returns a bool indicating if it can be casted to json or not.
    :returns: bool
    """
    try:
        int(obj)
        return True
    except:  # pylint: disable=bare-except
        return False


def is_json_serializable(obj):
    """
    is_json_serializable provided an object returns a bool indicating if it can be serialized to json or not.
    :returns: bool
    """
    try:
        json.dumps(obj)
        return True
    except:  # pylint: disable=bare-except
        return False



""" Base Model """
import json
from collections import OrderedDict
from typing import Dict, Any

import re
import simplejson
from schematics import Model
from schematics.exceptions import DataError

from wavelength.errors.exceptions import Base422Exception


class BaseModel(Model):
    """ Base class for deriving models (request && response) """

    def __init__(self, *args, **kwargs):
        try:
            kwargs['validate'] = kwargs.get('validate', True)
            super(BaseModel, self).__init__(*args, **kwargs)
        except (DataError, Exception) as error:
            raise Base422Exception(str(error)) from error

    def to_json(self):
        """ Camel-case and serialize dao to dict
        :return: dict
        """
        source = self.to_primitive()
        return BaseModel.convert_dict(source, BaseModel.to_camel_case)

    def to_dumps(self):
        """ Camel-case and serialize dao to JSON string
        :return: str
        """
        return json.dumps(self.to_json())

    def from_json(self, source):
        """  Builds dao from a camel-cased dict
        :param source: camel-cased dict
        :return: None
        """

        ctor = BaseModel.convert_dict(source, BaseModel.to_snake_case)
        self.__init__(ctor)
        return self

    def from_jsons(self, source):
        """ Builds dao from a JSON string
        :param source:
        :return: None
        """
        result = json.loads(source)
        self.from_json(result)
        return self

    @classmethod
    def convert_dict(cls, source, conversion):
        """  Recursively converts dict keys based on passed in conversion function
        :param self:
        :param source:
        :param conversion:
        :return:
        """
        result = OrderedDict()
        for k in source:
            key = conversion(k)
            if isinstance(source[k], dict):
                result[key] = BaseModel.convert_dict(source[k], conversion)
            elif isinstance(source[k], list):
                result[key] = BaseModel.convert_list(source[k], conversion)
            else:
                result[key] = source[k]
        return result

    @classmethod
    def convert_list(cls, source, conversion):
        """  Recursively converts list items based on passed in conversion function
        :param self:
        :param source:
        :param conversion:
        :return:
        """
        result = []
        for i in source:
            if isinstance(i, list):
                result.append(BaseModel.convert_list(i, conversion))
            elif isinstance(i, dict):
                result.append(BaseModel.convert_dict(i, conversion))
            else:
                result.append(i)
        return result

    @classmethod
    def to_camel_case(cls, word, uppercase_first_letter=False):
        """ Convert strings to CamelCase.
        @param word: str to convert
        @param uppercase_first_letter: boolean True if first letter is upper
        """
        if uppercase_first_letter:
            return re.sub(r"(?:^|_)(.)", lambda m: m.group(1).upper(), word)

        return word[0].lower() + BaseModel.to_camel_case(word, True)[1:]

    @classmethod
    def to_snake_case(cls, word):
        """ Convert strings to snake_case.
        @param word: str to convert
        """
        word = re.sub(r"([A-Z]+)([A-Z][a-z])", r'\1_\2', word)
        word = re.sub(r"([a-z\d])([A-Z])", r'\1_\2', word)
        word = word.replace("-", "_")
        return word.lower()


class SerializerBase:
    """
    Utility class to facilitate converting objects to dicts
    """

    def json_from_key(self, keys: list) -> Dict[str, Any]:
        """
        builds a dict from the supplied keys
        :param keys: list[str]
        :return: dict
        """
        result = {}  # type: Dict[str, Any]
        for key in keys:
            if hasattr(self, key):
                result.update(**{key: getattr(self, key)})
        return result

    def dumps_from(self, keys: list) -> str:
        """
        Converts to dict then serializes as to string
        :param keys: list[str]
        :return: str
        """
        result = self.json_from_key(keys)
        return simplejson.dumps(result)



"""
Module for supporting a hard delete on a pynamo item
"""
from wavelength.dal.dao.commands.pynamo_command import BasePynamoCommand
from wavelength.dal.pynamo_models.util import pynamo_write_handler
from wavelength.errors.exceptions import Base422Exception
from wavelength.model.util import get_posix_timestamp


class PynamoCreateCommand(BasePynamoCommand):
    """
    Creates a new record based on the parent model
    """

    def _execute(self):
        """
        Creates using retryable method
        :return: None
        """
        model = self._create_record()

        return self._parent.build(model.to_json())

    @pynamo_write_handler
    def _create_record(self):
        """
        Create record
        :return:None
        """
        try:
            now = get_posix_timestamp()
            self._parent.override_protected_field('created_at', now)
            self._parent.override_protected_field('modified_at', now)

            record = self._parent.db_model(*self._parent.get_key(),
                                           **self._parent.get_creation_actions())
            # TODO consider supporting conditions
            record.save()
            return record
        except ValueError as verr:
            raise Base422Exception('Model creation failed', reason=str(verr))

    def __call__(self):
        return self._execute()



"""
Whitelisting for logs
"""
FILTER_VALUE = "*"


def nested_get(input_dict, nested_keys):
    """
    Provided a dictionary and a list of strings (representing keys). This
    function uses the items wihtin nested_keys as keys within the input_dict.
    And will attempted to access and return the object outlined in the
    nested_keys list.
    """
    a_dict = input_dict
    for key in nested_keys:
        new_val = _nested_get(a_dict, key)
        if new_val is not None:
            a_dict = new_val
        else:
            return None
    return a_dict


def _nested_get(a_dict, key):
    new_val = None
    if hasattr(a_dict, 'get'):
        new_val = a_dict.get(key)

    return new_val


class NestedFilter:
    """
    NestedFilter is a data structure that is used for filtering out nested
    dictionaries. The path attribute is used to access a nested dictionary.
    The attrs attrs attribute is used to describe the attributes on the
    nested dictionary to modify.
    """

    def __init__(self, path, attrs):
        self.path = path
        self.attrs = attrs


def whitelist(input_dict, nested_filter, filter_value=FILTER_VALUE):
    """
    whitelist provided a dictionary, and a an instance of nested_filter.
    Will attempt to modify the nested input_dict as request.
    """
    nested_val = nested_get(input_dict, nested_filter.path)
    if (nested_val) and (isinstance(nested_val, dict)):
        return _transform_nested_whitelist(input_dict, nested_filter, filter_value)

    return input_dict


def _transform_nested_whitelist(input_dict, nested_filter, filter_value):
    length = len(nested_filter.path)
    a_dict = input_dict
    for i in range(length):
        k = nested_filter.path[i]
        if i == (length - 1):  # last item in nested_keys
            a_dict[k] = _whitelist_dict(a_dict.get(k), nested_filter.attrs, filter_value)
            break
        a_dict = a_dict.get(k)
    return input_dict


def _whitelist_dict(input_dict, attrs, filter_value):
    for k in input_dict:
        if k not in attrs:
            input_dict[k] = filter_value
    return input_dict



"""
Module to support loading a single record by key
"""
from typing import Any

from wavelength.dal.dao.commands.post_filter_pynamo_command import filter_deleted_items, PostFilteredPynamoCommand
from wavelength.dal.dao.models.platform_persistence_model import PynamoPersistenceModel


class PynamoGetByIdCommand(PostFilteredPynamoCommand):
    """
    This class encapsulates the updating of a versioned record
    """

    def __init__(self, parent_model: PynamoPersistenceModel, **kwargs):
        """

        :param parent_model: PynamoPersistenceModel (domain model object)
        :param post_filter: callable->bool function to run retrieved items against
        """
        filter_function = kwargs.get('post_filter', filter_deleted_items)

        super().__init__(parent_model, filter_function)
        self.keys = None

    def _execute(self) -> Any:
        """
        Run the command
        :return:Union[List[PynamoPersistenceModel],PynamoPersistenceModel]
        """
        result = super()._execute()
        self.keys = None

        return self._parent.build(result.to_json())

    def _get_key(self):
        return self.keys if self.keys else self._parent.get_key()

    def __call__(self, *keys):
        self.keys = keys
        return self._execute()



"""
Common utils for Dynamo interop
"""
import base64
import binascii
import json

from botocore.exceptions import ClientError
from retrying import retry

from wavelength.aws.boto_helper import is_a_boto3_throttle_error
from wavelength.errors.exceptions import Base409Exception, Base422Exception, Base429Exception
from wavelength.logging.slog import StructLog as log


class BotoThrottleConfig:
    """
    Class to set throttle retry values
    """
    max_retries = 5
    dynamo_db_timeout = 3.1
    max_write_backoff = 2500
    max_read_backoff = 2500
    wait_exponential_multiplier = 500


class OnThrottleError:
    """
    Boto3 Throttle handler for retry
    """

    def __init__(self, client_name):
        self._client_name = client_name

    def __call__(self, exception):
        """Return True if we should retry (in this case when it's an IOError), False otherwise"""
        if exception:
            throttle_error = is_a_boto3_throttle_error(exception)
            if throttle_error:
                log.warn(f'{self._client_name}_Operation',
                         'A throttle error was triggered',
                         operation='on_throttle_error',
                         data={'exception': str(exception)},
                         obj="exception",
                         exc_info=True)
            return throttle_error
        return False


def dynamo_read_handler(func):
    """ Retry dynamodb read wrapper, returns value from function or passes the exception upward """

    @retry(wait_exponential_multiplier=BotoThrottleConfig.wait_exponential_multiplier,
           wait_exponential_max=BotoThrottleConfig.max_read_backoff,
           stop_max_attempt_number=BotoThrottleConfig.max_retries,
           retry_on_exception=OnThrottleError('DynamonDB'))
    def wrapper(*args, **kwargs):
        """ Standard function wrapper """
        return func(*args, **kwargs)

    return wrapper


def dynamo_write_handler(func):
    """ Retry dynamodb write wrapper, returns value from function,
        true if there are concurrency issues or write fails, false if successful
        Non Client Error exception are not processed
        returns:
            True if a concurrency issues occurred
            False if process is successful
    """

    @retry(wait_exponential_multiplier=1000,
           wait_exponential_max=BotoThrottleConfig.max_write_backoff,
           stop_max_attempt_number=BotoThrottleConfig.max_retries,
           retry_on_exception=OnThrottleError('DynamonDB'))
    def wrapper(*args, **kwargs):
        """ Standard function wrapper """
        try:
            return func(*args, **kwargs)
        except ClientError as err:
            if err.response['Error']['Code'] == "ConditionalCheckFailedException":
                log.warn('CONDITIONAL_CHECK_FAILURE',
                         'Failed to update item due to conditional check failure',
                         operation=func.__name__, data={'exception': str(err)},
                         obj="exception", exc_info=True)
                raise Base409Exception(err)
            raise err

    return wrapper


class OnThrottleOrConditionalCheckError:
    """
    Boto3 Throttle handler for conditional check
    """

    def __init__(self, client_name):
        self._client_name = client_name

    def __call__(self, exception):
        """Return True if we should retry (in this case when it's an IOError), False otherwise"""
        if exception:
            throttle_error = OnThrottleError(self._client_name)
            if throttle_error(exception):
                return throttle_error
            if (isinstance(exception, ClientError) and
                    exception.response['Error']['Code'] == "ConditionalCheckFailedException"):
                log.warn('CONDITIONAL_CHECK_FAILURE',
                         'Failed to update item due to conditional check failure',
                         operation='_set_deleted', data={'exception': str(exception)},
                         obj="exception", exc_info=True)
                return True
        return False


def encode_start_token(token: dict) -> str:
    """
    Converts dict into base-64 encoded serialized JSON
    :param token: dict
    :return: str
    """
    if token is not None:
        return base64.b64encode(
            json.dumps(token).encode('utf-8')).decode('utf-8')
    return token


def decode_start_token(token: str = None):
    """
    Deserialize token back to dict
    :param token: str
    :return: dict
    """
    if token:
        try:
            return json.loads(base64.b64decode(token))
        except (binascii.Error, json.JSONDecodeError, UnicodeDecodeError) as berr:
            raise Base422Exception('Invalid startToken', reason=str(berr))

    return None


def validate_start_token(token: dict, keys: list):
    """
    Ensures that certain keys provided reside in the provided dictionary
    :param token: dict (representing decoded token)
    :param token: list (representing list of keys that should be in token)
    :return: None
    """
    if not isinstance(token, dict):
        raise Base422Exception('Invalid startToken', reason='start token is in an invalid format')

    for k in keys:
        if k not in token:
            reason = f"{k} is a required attribute that is not present in the provided start token"
            raise Base422Exception('Invalid startToken', reason=reason)


def decode_and_validate_start_token(token: str, keys: list):
    """
    Will attempt to decode the provided token. Which should be a dictionary. It will then attempt to ensure that the
    decoded dictionary matches the expected strucuture.
    :param token: str (representing an encoded token)
    :param token: list (representing list of keys that should be in token, once decoded)
    :return: None
    """
    decoded = decode_start_token(token)
    if decoded is None:
        raise Base422Exception('Invalid startToken', reason='Invalid format')

    validate_start_token(decoded, keys)
    return decoded


def handle_dynamodb_client_error(error):
    """
    Return correct application error from boto Client Error
    :param error: Exception
    :return: BODBaseException
    """
    error_map = {
        'ValidationException': Base422Exception,
        'ConditionalCheckFailedException': Base409Exception
    }
    if isinstance(error.cause, ClientError):
        code = error.cause.response['Error'].get('Code')
        message = error.cause.response['Error'].get('Message')
        exception = error_map.get(code)
        if exception:
            raise exception(error=message, reason=code)
        if is_a_boto3_throttle_error(error.cause):
            log.warn('EXCEPTION', error, error_code=code)
            raise Base429Exception(error.cause)


def check_version(arg1, arg2):
    """
    Version Equality Outcome, keeps code more concise and aides in testing
    :param arg1: Any
    :param arg2: Any
    :return: bool
    """
    if arg1 != arg2:
        raise Base409Exception('Model version conflict',
                               reason='Try getting the latest model and re-apply updates')



from os import path

from setuptools import setup, find_packages

# To use a consistent encoding
HERE = path.abspath(path.dirname(__file__))

# Get the long description from the README file
with open(path.join(HERE, 'README.rst')) as f:
    LONG_DESCRIPTION = f.read()

# Get the version number from the version.txt
with open(path.join(HERE, 'version.txt')) as f:
    VERSION = f.read()

with open(path.join(HERE, 'wavelength/requirements.txt')) as f:
    requires = f.read().splitlines()

setup(name='wavelength',
      version=VERSION,
      description='Common Utilities for Serverless Development',
      long_description=LONG_DESCRIPTION,
      classifiers=[
          'Programming Language :: Python :: 3'
      ],
      keywords='Lambda, Serverless, AWS, Dynamo, Structured Logging',
      author='Wavelength Serverless Team',
      author_email="dev@marrow-systems.com",
      packages=find_packages(exclude=['test*']),
      include_package_data=True,
      package_dir={'.': ''},
      install_requires=requires,
      zip_safe=False)



"""
Module to support updating a single record
"""

from wavelength.dal.dao.commands.base_update_pynamo_command import BaseUpdatePynamoCommand
from wavelength.dal.dao.commands.post_filter_pynamo_command import filter_deleted_items
from wavelength.dal.dao.models.platform_persistence_model import PynamoPersistenceModel


class PynamoUpdateCommand(BaseUpdatePynamoCommand):
    """
    This class encapsulates the updating of a versioned record
    """

    def __init__(self, parent_model: PynamoPersistenceModel):
        super().__init__(parent_model, filter_deleted_items)

    def __call__(self):
        return self._execute()



"""
Utilities for pynamo models
"""
from typing import List, Union

from pynamodb.exceptions import PynamoDBConnectionError
from pynamodb.indexes import Projection
from pynamodb.models import Model as PynamoModel
from pynamodb.pagination import ResultIterator
from retrying import retry

from wavelength.config import CONFIG
from wavelength.dal.dynamodb.util import encode_start_token, handle_dynamodb_client_error
from wavelength.errors.exceptions import Base429Exception, Base5xxException

PLATFORM_DB_PROPS = [
    'hash_id',
    'range_id',
    'kind',
]
COMMON_DB_PROPS = [

    'created_at',
    'modified_at',
    'table_state',
    'version'
]


class QueryResult:
    """
    Translates ResultIterator into a typed model and unwinds the internal generator
    """

    def __init__(self, result: ResultIterator):
        self._items = list(result)
        self._last_evaluated_key = encode_start_token(
            result.last_evaluated_key) if result.last_evaluated_key else None
        self._total_count = result.total_count

    @property
    def items(self) -> List[PynamoModel]:
        """
        Items returned from query
        :return: List[PynamoModel]
        """
        return self._items

    @property
    def last_evaluated_key(self) -> Union[str, dict, None]:
        """
        Next key token
        :return: Union[str, dict, None]
        """
        return self._last_evaluated_key

    @property
    def total_count(self) -> int:
        """
        Count from query
        :return: int
        """
        return self._total_count


class On429ThrottleError:
    """
    Boto3 Throttle handler for retry
    """

    def __init__(self, client_name):
        self._client_name = client_name

    def __call__(self, exception):
        """Return True if we should retry (in this case when it's an IOError), False otherwise"""
        return isinstance(exception, Base429Exception)


def pynamo_read_handler(func):
    """ Retry dynamodb read wrapper, returns value from function or passes the exception upward """

    @retry(wait_exponential_multiplier=CONFIG.CLIENT_RETRY_WAIT_EXPONENTIAL_MULTIPLIER,
           wait_exponential_max=CONFIG.CLIENT_RETRY_WAIT_EXPONENTIAL_MAX,
           stop_max_attempt_number=CONFIG.MAX_NUMBER_OF_CLIENT_RETRIES,
           retry_on_exception=On429ThrottleError('DynamonDB'))
    def wrapper(*args, **kwargs):
        """ Standard function wrapper """
        try:
            return func(*args, **kwargs)
        except PynamoDBConnectionError as err:
            handle_dynamodb_client_error(err)
            raise Base5xxException() from err

    return wrapper


def pynamo_write_handler(func):
    """ Retry dynamodb write wrapper, returns value from function or passes the exception upward """

    @retry(wait_exponential_multiplier=CONFIG.CLIENT_RETRY_WAIT_EXPONENTIAL_MULTIPLIER,
           wait_exponential_max=CONFIG.CLIENT_RETRY_WAIT_EXPONENTIAL_MAX,
           stop_max_attempt_number=CONFIG.MAX_NUMBER_OF_CLIENT_RETRIES,
           retry_on_exception=On429ThrottleError('DynamonDB'))
    def wrapper(*args, **kwargs):
        """ Standard function wrapper """
        try:
            return func(*args, **kwargs)
        except PynamoDBConnectionError as err:
            handle_dynamodb_client_error(err)
            raise Base5xxException() from err

    return wrapper


def get_pynamo_model_keys(db_model: PynamoModel):
    """
    Scan model attributes and pluck the keys
    :param db_model: PynamoModel
    :return: list
    """
    result = []
    for prop, val in getattr(db_model, '_attributes').items():

        if val.is_hash_key or val.is_range_key:
            result.append(prop)
    return result


class IndexBaseMeta:
    """
    Base Meta (useless, really as the actual props are set in serverless.yml but Pynamo needs them...)
    """
    projection = Projection()
    read_capacity_units = -1
    write_capacity_units = -1


class ModelBaseMeta:
    """
    Base Meta (useless, really as the actual props are set in serverless.yml but Pynamo needs them...)
    """
    region = CONFIG.AWS_REGION



"""
Module supporting soft delete behavior with a DynamoDB backend
"""
from typing import Dict

from pynamodb.expressions.operand import Path as PynamoAction

from wavelength.dal.dal_config import DalConfig
from wavelength.dal.dao.commands.base_update_pynamo_command import BaseUpdatePynamoCommand
from wavelength.dal.dao.commands.post_filter_pynamo_command import filter_deleted_items
from wavelength.dal.dao.models.platform_persistence_model import PynamoPersistenceModel


class PynamoSoftDeleteCommand(BaseUpdatePynamoCommand):
    """
    Class to encapsulate performing a soft delete command
    """

    def __init__(self, parent_model: PynamoPersistenceModel):
        super().__init__(parent_model, filter_deleted_items)

    def _execute(self):
        self._parent.override_protected_field('table_state', DalConfig.table_state_deleted)
        return super()._execute()

    def _build_model_update_actions(self) -> Dict[str, PynamoAction]:
        """
        Override to set table state == deleted
        :return: dict
        """

        return {
            **super()._build_model_update_actions(),
            **{
                'table_state': self._parent.db_model.table_state.set(DalConfig.table_state_deleted),
            }}

    def __call__(self):
        return self._execute()



"""
Base mode for cache enabled pynamo model access
"""
from cachetools import TTLCache
from pynamodb.models import Model as PynamoModel

from wavelength.config import CONFIG
from wavelength.dal.pynamo_models.util import QueryResult


class CachingPynamoModel(PynamoModel):
    """
    Overrides a few methods on the Model class to enable selective caching
    """

    cache = TTLCache(maxsize=128, ttl=CONFIG.DATA_CACHE_TTL)

    @classmethod
    def query(cls, *args, **kwargs) -> QueryResult:  # type: ignore  # pylint: disable=arguments-differ
        """
        Override base for caching behavior
        :param hash_key: str
        :param range_key_condition: str
        :param filter_condition:  Condition
        :param consistent_read: bool
        :param index_name: str
        :param scan_index_forward: bool
        :param conditional_operator: List
        :param limit: int
        :param last_evaluated_key: Union[str, dict, None]
        :param attributes_to_get: List
        :param page_size: int
        :param rate_limit: int
        :param filters: List
        :return: QueryResult
        """
        result = QueryResult(super().query(*args, **kwargs))
        for item in result.items:
            cls._put_model_cache(item)
        return result

    @classmethod
    def get(cls,
            hash_key,
            range_key=None,
            consistent_read=False,
            attributes_to_get=None):
        """
        Returns a single object using the provided keys
        :param hash_key: str
        :param range_key: str
        :param consistent_read: bool
        :param attributes_to_get: List
        :return: PynamoModel
        """
        key = cls.get_cache_key(hash_key, range_key)
        result = cls.cache.get(key)

        if not result:
            result = super().get(hash_key, range_key, consistent_read, attributes_to_get)
            cls.cache[key] = result

        return result

    def save(self, condition=None, conditional_operator=None, **expected_values):
        """
        Save this object to dynamodb
        :param condition: Condition
        :param conditional_operator: List
        :param expected_values: List
        :return: dict
        """
        result = super().save(condition, conditional_operator, **expected_values)
        self._put_model_cache(self)
        return result

    def update(self, attributes=None, actions=None, condition=None, conditional_operator=None, **expected_values):
        """
        Updates an item using the UpdateItem operation.
        :param attributes: List
        :param actions: List
        :param condition: Condition
        :param conditional_operator: List
        :param expected_values: List
        :return: dict
        """
        result = super().update(attributes, actions, condition, conditional_operator, **expected_values)
        self._put_model_cache(self)
        return result

    def delete(self, condition=None, conditional_operator=None, **expected_values):
        """
        Deletes this object from dynamodb
        :param condition: Condition
        :param conditional_operator: List
        :param expected_values: List
        :return: dict
        """
        result = super().delete(condition, conditional_operator, **expected_values)
        self._delete_model_cache(self)
        return result

    @classmethod
    def get_cache_key(cls, hash_key, range_key=None):
        """
        Builds hash key from parameters
        :param args: positionals
        :param kwargs: keywords
        :return: str
        """
        result = hash_key
        if range_key:
            result = f'{hash_key}:::{range_key}'

        return result

    @classmethod
    def _put_model_cache(cls, model: PynamoModel):
        """
        Replaces an item in the cache based on its own keys
        :param model: PynamoModel
        :return: None
        """
        key = cls.get_cache_key(*getattr(model, '_get_keys')())
        cls.cache[key] = model

    @classmethod
    def _delete_model_cache(cls, model: PynamoModel):
        """
        Replaces an item in the cache based on its own keys
        :param model: PynamoModel
        :return: None
        """
        key = cls.get_cache_key(*getattr(model, '_get_keys')())
        cls.cache.pop(key)
